'''
#Fig. 3
# import packages
from pyDHM import utilities
from pyDHM import phaseShifting

# Load the holograms
inp0 = utilities.imageRead('holo1.png')
inp1 = utilities.imageRead('holo2.png')
inp2 = utilities.imageRead('holo3.png')
inp3 = utilities.imageRead('holo4.png')
inp4 = utilities.imageRead('holo5.png')

# Phase shifting using the PS strategies
output = phaseShifting.PS5(inp0, inp1, inp2, inp3, inp4)
output = phaseShifting.PS4(inp0, inp1, inp2, inp3)
output = phaseShifting.PS3(inp0, inp1, inp2)

# Display the phase reconstruction
phase = utilities.phase(output)
utilities.imageShow(output, 'Phase reconstruction') 
'''

'''
#Fig. 4
# import packages 
from pyDHM import utilities
from pyDHM import phaseShifting 

# Load the holograms
inp0 = utilities.imageRead('holo1.png')
inp1 = utilities.imageRead('holo2.png')
inp2 = utilities.imageRead('holo3.png')
inp3 = utilities.imageRead('holo4.png')

# Phase shifting via SOSR, BPS3 or BPS2
output = phaseShifting.SOSR(inp0, inp1, inp2, inp3, 633e-9, 6.9e-6, 6.9e-6, 1, 4)
output = phaseShifting.BPS3(inp0, inp1, inp2, 0.532, 2.9, 2.9)
output = phaseShifting.BPS2(inp0, inp1, 0.532, 2.9, 2.9)

# Display the phase reconstruction
phase = utilities.phase(output)
utilities.imageShow(phase, 'Phase reconstruction')
'''

'''
#Fig. 5
# import packages 
from pyDHM import utilities
from pyDHM import phaseCompensation

# Load the hologram
inp = utilities.imageRead('hologram.jpg')

# Compensation using the available approaches
output = phaseCompensation.FRS(inp, True, 0.633, 6.9, 6.9, 2, 10)
output = phaseCompensation.ERS(inp, True, 0.633, 6.9, 6.9, 5, 0.2)
output = phaseCompensation.CFS(inp, 0.532, 2.6, 2.6)

# Display the phase reconstruction
phase = utilities.phase(output)
utilities.imageShow(phase, 'Phase reconstruction')
'''

'''
#Fig. 6 
# import packages 
from pyDHM import utilities
from pyDHM import phaseCompensation

# Load the hologram
inp = utilities.imageRead('hologram.tif')

# FT of the hologram
ft_holo = utilities.FT(inp)
ft_holo = utilities.intensity(ft_holo, True)
utilities.imageShow(ft_holo, 'FT hologram')

# Numerical compensation using the CNT approach
output = phaseCompensation.CNT(inp, 0.633, 6.9, 6.9, spatialFilter='sfmr')

# Display the phase reconstruction
phase = utilities.phase(output)
utilities.imageShow(phase, 'Phase reconstruction')
'''

'''
#Fig 7
# import packages 
from pyDHM import utilities
from pyDHM import numericalPropagation

# Load the hologram
hologram =  utilities.imageRead('hologram.jpg')
utilities.imageShow(hologram, 'Hologram')

# FT of the hologram
ft_holo = utilities.FT(hologram)
ft_holo = utilities.intensity(ft_holo, True)
utilities.imageShow(ft_holo, 'FT hologram')

# Circular spatial filter
filter = utilities.sfc(hologram, 160, 303, 276, True)

# Numerical propagation using the angular spectrum
output = numericalPropagation.angularSpectrum(filter, 0.02, 0.633, 6.9, 6.9)

# Display the output field 
intensity = utilities.intensity(output, False)
utilities.imageShow(intensity,'ouput field')
'''

'''
#Fig. 8
# import packages 
from pyDHM import utilities
from pyDHM import numericalPropagation

# Load the hologram
input = utilities.imageRead('hologram.tif')
utilities.imageShow(input, 'input field')

# FT of the hologram
ft_holo = utilities.FT(input)
ft_holo = utilities.intensity(ft_holo, True)
utilities.imageShow(ft_holo, 'FT hologram')

# Spatial filter
filter = utilities.sfr(input, 280, 500, 150, 340, True)

# Numerical propagation using the Fresnel transforms
output = numericalPropagation.fresnel(input, -450, 0.000633, 0.005, 0.005)
output =  numericalPropagation.bluestein(input, 0.03, 633e-9, 7.4e-6, 7.4e-6, 1.48e-5, 1.48e-5)  

# Display the output field
amplitude = utilities.amplitude(output, True)
utilities.imageShow(amplitude, 'output field')
'''



"""
Title-->            Phase phaseShifting package
Author-->           Ana Doblas, Carlos Trujillo, Raul Castaneda,
Date-->             21/02/2021
Last modified-->    11/04/2022
Groups-->           University of Memphis -  Optical Imaging Research laboratory (OIRL)
                    EAFIT University - Applied Optics Group
Abstract -->        Script for implementing six phase shifting algorithms implemented in pyDHM:
                    SOSR, BPS2, BPS3, PS5, PS4 and PS3.
Links-->            More information can be found on the website
                    https://github.com/catrujilla/pyDHM
"""

# libraries
import numpy as np
import math
from math import pi
from scipy.optimize import minimize
import scipy
import sys
import cv2

def SOSR(inp0, inp1, inp2, inp3, upper, wavelength, dx, dy, s=1, steps=4):
    '''
    # Function to recover the phase information of a sample from 4 DHM slightly off-axis acquisitions (quadrature Phase Shifting method).
    # The idea was first proposed by Ferraro, we upgraded the method to accurately calculate the synthetic reference wave
    # to obtain a fully compensated phase maps.
    # Inputs:
    # inpX - The input intensities (captured) pi/2 phase-shifted holograms
    # upper - Boolean variable defining whether the brightest point will be that on the upper part of the DHM hologram spectrum.
    # wavelength - Wavelength of the illumination source to register the DHM hologram
    # dx, dy - Pixel dimensions of the camera sensor used for recording the holograms
    '''

    if steps < s:
        print('Please, Enter a s value smaller than step')
        sys.exit()

    # Retrieving the input shape

    inp0 = inp0 - np.average(inp0)
    inp1 = inp1 - np.average(inp1)
    inp2 = inp2 - np.average(inp2)
    inp3 = inp3 - np.average(inp3)
    M, N = inp0.shape

    # Creating a mesh_grid to operate in world-coordinates
    x = np.arange(0, N, 1)  # array x
    y = np.arange(0, M, 1)  # array y
    X, Y = np.meshgrid(x - (N / 2), y - (M / 2), indexing='xy')  # meshgrid XY

    # Phase Shifting process

    # Location of the brightest point in the first phase-shifted DHM hologram spectrum
    # First, the need to Fourier transform and biuld the power spectrum
    field_spec = np.fft.fftshift(inp0)
    field_spec = np.fft.fft2(field_spec)
    field_spec = np.fft.fftshift(field_spec)
    pow_espec = intensity(field_spec, True)

    # Defining width and height of point of maxima in the ROI determination
    # 10% is estimated to be the proper distance to separated an in-line hologram from an off-axis one in the Spectrum
    M_search = int(M - (M * 0.1))
    N_search = int(N - (N * 0.1))
    # print (M_search)

    # Determining the point of maxima in the power spectrum
    if (upper == True):
        # Find the location of the maxima point in the upper part of the DHM hologram spectrum
        indI = np.unravel_index(np.argmax(pow_espec[1:(int(M / 2) - M_search), 1:N], axis=None),
                                pow_espec[1:(int(M / 2) - M_search), 1:N].shape)
        # print (indI)
    else:
        # Find the location of the maxima point in the lower part of the DHM hologram spectrum
        indI = np.unravel_index(np.argmax(pow_espec[1:M, 1:(int(N / 2) - N_search)], axis=None),
                                pow_espec[1:M, 1:(int(N / 2) - N_search)].shape)
        # print (indI)

    # DC coordinates (centers of the DHM hologram Spectrums)
    fx_0, fy_0 = N / 2, M / 2

    # Coordinates of maxima points in the DHM hologram spectrums
    fx_1, fy_1 = indI[1] + 1, indI[0] + 1  # +1 is the correction of the offset added by numpy?

    # print (fx_0, fy_0, fx_1, fy_1)

    sum_max = 0
    arrayX = np.linspace(fx_1 - s, fx_1 + s, steps)
    arrayY = np.linspace(fy_1 - s, fy_1 + s, steps)

    k = (2 * pi) / wavelength

    print("Synthetic building of the reference wavefront started....")

    for fx_temp in arrayX:
        for fy_temp in arrayY:
            # print (fx_temp, fy_temp)
            theta_x = math.asin((fx_0 - fx_temp) * wavelength / (N * dx))
            theta_y = math.asin((fy_0 - fy_temp) * wavelength / (M * dy))

            R0 = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
            R1 = 1j * np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
            R2 = (-1) * np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
            R3 = (-1j) * np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))

            # Four-Quarter Phase Shifting (Ferraro's proposal)
            Isynth = (R0 * inp0) + (R1 * inp1) + (R2 * inp2) + (R3 * inp3)

            # module 2-pi phase retrieval
            phase = np.angle(Isynth)

            # Thresholding process
            minVal = np.amin(phase)
            maxVal = np.amax(phase)
            phase_sca = (phase - minVal) / (maxVal - minVal)
            binary_phase = (phase_sca < 0.1)

            # plt.imshow(phase, cmap='gray'), plt.title("phase")  # image in gray scale
            # plt.show()  # show image

            # Applying the summation and thresholding metric
            sum = np.sum(np.sum(binary_phase))
            if sum > sum_max:
                x_max_out = fx_temp;
                y_max_out = fy_temp;
                sum_max = sum;
            # print (sum)

    # print(x_max_out, y_max_out)
    print("Synthetic reference wavefront built.")

    # Retrieving the best reconstruction (compensated phase)
    theta_x = math.asin((fx_0 - x_max_out) * wavelength / (N * dx))
    theta_y = math.asin((fy_0 - y_max_out) * wavelength / (M * dy))
    R0 = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    R1 = 1j * np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    R2 = (-1) * np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    R3 = (-1j) * np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))

    # Four-Quarter Phase Shifting (Ferraro's proposal)
    Isynth = (R0 * inp0) + (R1 * inp1) + (R2 * inp2) + (R3 * inp3)

    comp_phase = Isynth

    print("Phase-shifting reconstruction finished.")

    return comp_phase


def BPS2(Inp0, Inp1, wavelength, dx, dy):
    '''
    # Function to recover the phase information of a sample from two slightly off-axis acquisitions.
    # Inputs:
    # InpX - The input intensities
    # wavelength - Wavelength of the illumination source to register the DHM hologram
    # dx, dy- Pixel dimensions of the camera sensor used for recording the holograms
    '''

    if scipy.__version__ == None:
        print('Please install scipy library and import the function minimize')
        sys.exit()

    # retrieving the input shape
    inp = np.array(Inp0)
    M, N = inp.shape

    # FT hologram
    field_spec = np.fft.fft2(inp)
    field_spec = np.fft.fftshift(field_spec)
    pow_espec = intensity(field_spec, True)

    # finding the max peaks for +1 order in I or II quadrant
    height_half = round(int(M / 2))
    mask = np.zeros((M, N))
    mask[0:height_half - 1, 0:N] = 1
    field_spec_tem = pow_espec * mask
    maximum = np.amax(field_spec_tem)
    fy_max_d1, fx_max_d1 = np.where(field_spec_tem == maximum)

    # finding the max peaks for -1 order in III or IV quadrant
    height_half = round(int(M / 2))
    mask = np.zeros((M, N))
    mask[height_half + 1:M, 0:N] = 1
    field_spec_tem = pow_espec * mask
    maximum = np.amax(field_spec_tem)
    fy_max_d2, fx_max_d2 = np.where(field_spec_tem == maximum)

    # creating seeds for the minimization
    test_theta = np.random.randint(0, 360, (1, 2))
    test_theta = test_theta * math.pi / 180
    test_theta = [test_theta[0, 0], test_theta[0, 1]]

    # minimization
    print("Minimization process started.....")
    res = minimize(costFunction2, test_theta, args=(Inp0, Inp1, fx_max_d1, fy_max_d1, fx_max_d2,
                                                        fy_max_d2), method='Cobyla', tol=1e-6)
    print("Minimization process finished.")

    print("Phase compensation started....")
    # computing d1 and d2
    x = res.x
    theta1 = x[0]
    theta2 = x[1]
    ones = np.ones((2, 2))
    Matrix = [[1, np.exp(1j*theta1)],
         [1, np.exp(1j*theta2)]]
    Matrix = Matrix * ones
    Minv = np.linalg.inv(Matrix)
    d2 = Minv[1, 0]*Inp0 + Minv[1, 1]*Inp1

    # FFT d2
    FTd2 = np.fft.fft2(d2)
    FTd2 = np.fft.fftshift(FTd2)

    # spatial filtering
    out, fx_max, fy_max = spatialFiltering(FTd2, M, N)

    # compensation process
    # computing the digital reference wave
    ref_wave = referenceWave(M, N, wavelength, dx, dy, fx_max, fy_max)
    comp_phase = out * ref_wave
    # introducir metodo de compensación
    print("Phase compensation finished.")

    return comp_phase


def BPS3(Inp0, Inp1, Inp2, wavelength, dx, dy):
    '''
    # Function to recover the phase information of a sample from three slightly off-axis acquisitions.
    # Inputs:
    # InpX - The input intensities (captured)
    # wavelength - Wavelength of the illumination source to register the DHM hologram
    # dx, dy - Pixel dimensions of the camera sensor used for recording the hologram
    '''
    
    if scipy.__version__ == None:
        print('Please install scipy library and import the function minimize')
        sys.exit()


    # Retrieving the input shape
    inp = np.array(Inp0)
    M, N = inp.shape

    # FT hologram
    field_spec = np.fft.fft2(inp)
    field_spec = np.fft.fftshift(field_spec)
    pow_espec = intensity(field_spec, True)

    # finding the max peaks for +1 order in I or II quadrant
    height_half = round(int(M / 2))
    mask = np.zeros((M, N))
    mask[0:height_half - 1, 0:N] = 1
    field_spec_tem = pow_espec * mask
    maximum = np.amax(field_spec_tem)
    fy_max_d1, fx_max_d1 = np.where(field_spec_tem == maximum)

    # finding the max peaks for -1 order in III or IV quadrant
    height_half = round(int(M / 2))
    mask = np.zeros((M, N))
    mask[height_half + 1:M, 0:N] = 1
    field_spec_tem = pow_espec * mask
    maximum = np.amax(field_spec_tem)
    fy_max_d2, fx_max_d2 = np.where(field_spec_tem == maximum)

    # creating seeds for the minimization
    test_theta = np.random.randint(0, 360, (1, 3))
    test_theta = test_theta * math.pi / 180
    test_theta = [test_theta[0, 0], test_theta[0, 1], test_theta[0, 2]]

    # minimization
    print("Minimization process started.....")
    res = minimize(costFunction3, test_theta, args=(Inp0, Inp1, Inp2, fx_max_d1, fy_max_d1, fx_max_d2,
                                                    fy_max_d2), method='Cobyla', tol=1e-6)
    print("Minimization process finished.")

    print("Phase compensation started....")
    # computing d1, d2 and d3
    x = res.x
    theta1 = x[0]
    theta2 = x[1]
    theta3 = x[2]
    ones = np.ones((3, 3))
    Matrix = [[1, np.exp(1j * theta1), np.exp(-1j * theta1)],
         [1, np.exp(1j * theta2), np.exp(-1j * theta2)],
         [1, np.exp(1j * theta3), np.exp(-1j * theta3)]]
    Matrix = Matrix * ones
    Minv = np.linalg.inv(Matrix)
    d = Minv[1, 0] * Inp0 + Minv[1, 1] * Inp1 + Minv[1, 2] * Inp2


    # FFT d
    FTd = np.fft.fft2(d)
    FTd = np.fft.fftshift(FTd)

    # spatial filtering
    out, fx_max, fy_max = spatialFiltering(FTd, M, N)

    # compensation process
    # computing the digital reference wave
    ref_wave = referenceWave(M, N, wavelength, dx, dy, fx_max, fy_max)
    comp_phase = out * ref_wave
    print("Phase compensation finished.")

    return comp_phase


def PS5(Inp0, Inp1, Inp2, Inp3, Inp4):
    '''
    # Function to recover the phase information of a sample from five DHM in-axis acquisitions holograms
    # Inputs:
    # inpX - The input intensities (captured) pi/2 phase-shifted holograms
    '''
    
    # determine if the hologram is on-axis
    orders, thresh = regime(Inp0)
    if orders != 1:
        print('PS5 algorithm requires on-axis holograms')
        sys.exit()

    # Retrieving the input shape
    inp0 = np.array(Inp0)
    inp1 = np.array(Inp1)
    inp2 = np.array(Inp2)
    inp3 = np.array(Inp3)
    inp4 = np.array(Inp4)

    print("Phase-shifing reconstruction started....")
    # computing the compensation
    comp_phase = np.arctan((2*(inp3-inp1))/(2*inp2-inp0-inp4))
    print("Phase-shifing reconstruction finished.")

    return comp_phase


def PS4(Inp0, Inp1, Inp2, Inp3):
    '''
    # Function to recover the phase information of a sample from four DHM in-axis acquisitions holograms
    # Inputs:
    # inpX - The input intensities (captured) pi/2 phase-shifted holograms
    '''
    
    # determine if the hologram is on-axis
    orders, thresh = regime(Inp0)
    if orders != 1:
        print('PS4 algorithm requires on-axis holograms')
        sys.exit()

    # Retrieving the input shape
    inp0 = np.array(Inp0)
    inp1 = np.array(Inp1)
    inp2 = np.array(Inp2)
    inp3 = np.array(Inp3)


    # compensation process
    print("Phase-shifing reconstruction started....")
    # computing the compensation
    comp_phase = np.arctan((inp3-inp1)/(inp2-inp0))
    print("Phase-shifing reconstruction finished.")

    return comp_phase


def PS3(Inp0, Inp1, Inp2):
    '''
    # Function to recover the phase information of a sample from three DHM in-axis acquisitions holograms
    # Inputs:
    # inpX - The input intensities (captured) pi/3 phase-shifted holograms
    '''

    '''
    # determine if the hologram is on-axis
    orders, thresh = regime(Inp0)
    if orders != 1:
        print('PS3 algorithm requires on-axis holograms ')
        sys.exit()
    '''
    
    # Retrieving the input shape
    inp0 = np.array(Inp0)
    inp1 = np.array(Inp1)
    inp2 = np.array(Inp2)

    print("Phase-shifing reconstruction started....")

    # computing the compensation
    comp_phase = np.arctan((np.sqrt(3)*(inp2-inp0))/((inp2+inp0)-(2*inp1)))
    print("Phase-shifing reconstruction finished.")

    return comp_phase
    
    
'''
Auxiliary functions
'''
# Function to determine if the holograms is off-axis or not
def regime(inp):
    holoFT = np.float32(inp)
    fft_holo = cv2.dft(holoFT, flags=cv2.DFT_COMPLEX_OUTPUT)
    fft_holo = np.fft.fftshift(fft_holo)
    fft_holo_image = 20 * np.log(cv2.magnitude(fft_holo[:, :, 0], fft_holo[:, :, 1]))
    minVal = np.amin(np.abs(fft_holo_image))
    maxVal = np.amax(np.abs(fft_holo_image))
    fft_holo_image = cv2.convertScaleAbs(fft_holo_image, alpha=255.0 / (maxVal - minVal),
                                         beta=-minVal * 255.0 / (maxVal - minVal))

    # cv2.imshow('Binary image_resize', fft_holo_image)
    # cv2.waitKey(0)

    # apply binary thresholding
    ret, thresh = cv2.threshold(fft_holo_image, 240, 255, cv2.THRESH_BINARY)
    #cv2.imshow('Binary image', thresh)
    thresh_rize = cv2.resize(thresh, (1024, 1024))
    #cv2.imshow('Binary image_resize', thresh_rize)
    #cv2.waitKey(0)


    contours, hierarchy = cv2.findContours(image=thresh_rize, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_SIMPLE)
    # draw contours on the original image
    fft_holo_image = cv2.resize(thresh, (1024, 1024))
    image_copy = fft_holo_image.copy()
    cv2.drawContours(image=image_copy, contours=contours, contourIdx=-1, color=(0, 255, 0), thickness=2,
                     lineType=cv2.LINE_AA)
    # cv2.imshow('None approximation', image_copy)
    # cv2.waitKey(0)
    orders = len(contours)
    # print(orders)
    return orders, thresh


# Function to calculate the intensity representation of a given complex field
def intensity(inp, log):
    # Function to calcule the intensity representation of a given complex field
    # Inputs:
    # inp - The input complex field
    # log - boolean variable to determine if a log representation is applied
    out = np.abs(inp)
    out = out * out

    if log == True:
        out = 20 * np.log(out)

    return out


# cost function for the BPS2 implementation
def costFunction2(test_theta, hologram1, hologram2, fx_max_d1, fy_max_d1, fx_max_d2, fy_max_d2):
    theta1 = test_theta[0]
    theta2 = test_theta[1]
    ones = np.ones((2, 2))
    Matrix = [[1, np.exp(1j * theta1)],
         [1, np.exp(1j * theta2)]]
    Matrix = Matrix * ones
    Minv = np.linalg.inv(Matrix)
    d1 = Minv[0, 0] * hologram1 + Minv[0, 1] * hologram2

    # FFT d1
    FTd1 = np.fft.fft2(d1)
    FTd1 = np.fft.fftshift(FTd1)

    # calculate J
    J = 1 - (np.abs(FTd1[fx_max_d1, fy_max_d1]) - np.abs(FTd1[fx_max_d2, fy_max_d2])) / \
        (np.abs(FTd1[fx_max_d2, fy_max_d2]) + np.abs(FTd1[fx_max_d1, fy_max_d1]))
    J = J[0]
    return J


# cost function for the BPS3 implementation
def costFunction3(test_theta, hologram1, hologram2, hologram3, fx_max_d1, fy_max_d1, fx_max_d2, fy_max_d2):
    theta1 = test_theta[0]
    theta2 = test_theta[1]
    theta3 = test_theta[2]
    ones = np.ones((3, 3))
    Matrix = [[1, np.exp(1j*theta1), np.exp(-1j*theta1)],
         [1, np.exp(1j*theta2), np.exp(-1j*theta2)],
         [1, np.exp(1j*theta3), np.exp(-1j*theta3)]]
    Matrix = Matrix * ones
    Minv = np.linalg.inv(Matrix)
    d3 = Minv[2, 0]*hologram1 + Minv[2, 1]*hologram2 + Minv[2, 2]*hologram3

    # FFT d3
    FTd3 = np.fft.fft2(d3)
    FTd3 = np.fft.fftshift(FTd3)

    # calculate J
    J = 1 + (1 - (np.abs(FTd3[fx_max_d2, fy_max_d2]) - np.abs(FTd3[fx_max_d1, fy_max_d1])) /
             (np.abs(FTd3[fx_max_d2, fy_max_d2]) + np.abs(FTd3[fx_max_d1, fy_max_d1])))
    J = J[0]
    return J


# Spatial filtering process for blind phase-shifting methods
def spatialFiltering(inp, M, N):
    # Find the max peaks I or II quadrant
    height_half = round(int(M / 2))
    mask = np.zeros((M, N))
    mask[0:height_half - 1, 0:N] = 1
    field_spec_tem = inp * mask
    maximum = np.amax(field_spec_tem)
    fy_max, fx_max = np.where(field_spec_tem == maximum)

    # Determination of the ROI size. To do this, we use the theoretical size of the +1 or -1 diffraction orders according...
    # ... to the distance of their centers to the DC coordinates (middle point in the DHM hologram spectrum).
    # d = np.sqrt(np.power(fx_max - M / 2, 2) + np.power(fy_max - N / 2, 2))
    # radius = d / 3
    radius = 100

    # Mask building
    mask = np.zeros((M, N), dtype=int)
    for p in range(0, N):
        for q in range(0, M):
            if np.sqrt((p - fy_max) ** 2 + (q - fx_max) ** 2) < radius:
                mask[p, q] = 1

    # Filtering the hologram
    tmp = inp * mask

    # Coming back to spatial domain (retrieving filtered hologram)
    out = np.fft.ifftshift(tmp)
    out = np.fft.ifft2(out)

    return out, fx_max, fy_max


# function to create the reference wave
def referenceWave(M, N, wavelength, dx, dy, fx_max, fy_max):
    x = np.arange(0, M, 1)  # array x
    y = np.arange(0, N, 1)  # array y
    X, Y = np.meshgrid(x - (M / 2), y - (N / 2), indexing='xy')  # meshgrid XY
    fx_0 = M / 2
    fy_0 = N / 2
    k = (2 * math.pi) / wavelength
    theta_x = math.asin((fx_0 - fx_max) * wavelength / (M * dx))
    theta_y = math.asin((fy_0 - fy_max) * wavelength / (N * dy))
    ref_wave = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    return ref_wave



"""
Title-->            Phase compensators package
Author-->           Carlos Trujillo, Ana Doblas and Raul Castaneda
Date-->             21/06/2021
Last modified-->    05/09/2022
Groups-->           University of Memphis -  Optical Imaging Research laboratory (OIRL)
                    EAFIT University - Applied Optics Group
Abstract -->        Script for implementing four-phase compensators algorithms implemented in pyDHM:
                    Full ROI Search (FRS), Efficient ROI Search (ERS), the Cost Function ROI (CFS),
                    and the compensation No-tele (CNT).
Links-->            More information can be found on the website
                    https://github.com/catrujilla/pyDHM
information -->     This package uses the python library scipy for the CFS function. For the scipy
                    installation please visit the webpage https://scipy.org/install/
"""

# libraries
import numpy as np
import math
from matplotlib import pyplot as plt
from math import pi
from scipy.optimize import minimize
import scipy
import sys
import cv2

def FRS(inp, upper, wavelength, dx, dy, s=2, step=10):
    '''
    # Function to compensate phase maps of off-axis DHM via the full ROI search algorithm.
    # Inputs:
    # inp - The input intensity (captured) hologram
    # upper - Boolean variable defining whether the ROI will be that on the upper part of the DHM hologram spectrum.
    # wavelength - Wavelength of the illumination source to register the DHM hologram
    # dx, dy - Pixel dimensions of the camera sensor used for recording the hologram
    # s = 2 and steps = 10
    '''
    
    if step < s:
        print('Please, Enter a s value smaller than step')
        sys.exit()

    # determine if the hologram is off-axis
    orders, thresh = regime(inp)
    if orders < 3:
        print('FRS algorithm requires off-axis holograms')
        sys.exit()
        

    # Retrieving the input shape
    inp = inp - np.average(inp)
    inp = np.array(inp)
    M, N = inp.shape

    # Creating a mesh_grid to operate in world-coordinates
    x = np.arange(0, N, 1)  # array x
    y = np.arange(0, M, 1)  # array y
    X, Y = np.meshgrid(x - (N / 2), y - (M / 2), indexing='xy')  # meshgrid XY

    # The spatial filtering process is executed
    print("Spatial filtering process started.....")
    holo_filter, indI, tmp = spatialFiltering(inp, M, N, upper)
    print("Spatial filtering process finished.")

    # DC coordinates (center of DHM hologram Spectrum)
    fx_0, fy_0 = N / 2, M / 2

    # Coordintes of maxima in the DHM hologram spectrum
    fx_1, fy_1 = indI[1] + 1, indI[0] + 1  # +1 is the correction of the offset added by numpy?

    sum_max = 0
    arrayX = np.linspace(fx_1 - s, fx_1 + s, step)
    arrayY = np.linspace(fy_1 - s, fy_1 + s, step)

    k = (2 * pi) / wavelength

    print("Phase compensation started....")

    for fx_temp in arrayX:
        for fy_temp in arrayY:
            # print (fx_temp, fy_temp)
            theta_x = math.asin((fx_0 - fx_temp) * wavelength / (N * dx))
            theta_y = math.asin((fy_0 - fy_temp) * wavelength / (M * dy))
            ref_wave = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
            # Compensation of the tilting angle for the off-axis acquisition
            reconstruction = holo_filter * ref_wave
            # module 2-pi phase retrieval
            phase = np.angle(reconstruction)
            # Thresholding process
            minVal = np.amin(phase)
            maxVal = np.amax(phase)
            phase_sca = (phase - minVal) / (maxVal - minVal)
            binary_phase = (phase_sca > 0.2)

            # plt.imshow(binary_phase, cmap='gray'), plt.title("binary_phase")  # image in gray scale
            # plt.show()  # show image

            # Applying the summation and thresholding metric
            sum = np.sum(np.sum(binary_phase))
            if sum > sum_max:
                x_max_out = fx_temp;
                y_max_out = fy_temp;
                sum_max = sum;
            # print (sum)

    # Retrieving the best reconstruction (compensated phase)
    theta_x = math.asin((fx_0 - x_max_out) * wavelength / (N * dx))
    theta_y = math.asin((fy_0 - y_max_out) * wavelength / (M * dy))
    ref_wave = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    comp_phase = holo_filter * ref_wave

    print("Phase compensation finished.")
    # print(x_max_out, y_max_out)

    return comp_phase


def ERS(inp, upper, wavelength, dx, dy, s, step):
    '''
    # Function to compensate phase maps of off-axis DHM via the fast ROI search algorithm.
    # Inputs:
    # inp - The input intensity (captured) hologram
    # upper - Boolean variable defining whether the ROI will be that on the upper part of the DHM hologram spectrum.
    # wavelength - Wavelength of the illumination source to register the DHM hologram
    # dx, dy - Pixel dimensions of the camera sensor used for recording the hologram
    # s = 5 and step = 0.2
    '''
    if step > s:
        print('Please, Enter a step value smaller than s')
        sys.exit()

    # determine if the hologram is off-axis
    orders, thresh = regime(inp)
    if orders < 3:
        print('ERS algorithm requires off-axis holograms')
        sys.exit()

    # Retrieving the input shape
    inp = inp - np.average(inp)
    inp = np.array(inp)
    M, N = inp.shape

    # Creating a mesh_grid to operate in world-coordinates
    x = np.arange(0, N, 1)  # array x
    y = np.arange(0, M, 1)  # array y
    X, Y = np.meshgrid(x - (N / 2), y - (M / 2), indexing='xy')  # meshgrid XY

    # The spatial filtering process is executed
    print("Spatial filtering process started.....")
    holo_filter, indI, tmp = spatialFiltering(inp, M, N, upper)
    print("Spatial filtering process finished.")

    # out = phc.intensity(tmp, True)
    # plt.imshow(out, cmap='gray'), plt.title("holo_filterFreq")  # image in gray scale
    # plt.show()  # show image

    # Phase Compensation process

    # DC coordinates (center of DHM hologram Spectrum)
    fx_0, fy_0 = N / 2, M / 2

    # Coordintes of maxima in the DHM hologram spectrum
    fx_1, fy_1 = indI[1] + 1, indI[0] + 1  # +1 is the correction of the offset added by numpy?

    # print (fx_0, fy_0, fx_1, fy_1)

    print("Phase compensation started....")

    k = (2 * pi) / wavelength

    fin = 0
    sum_max = 0

    fx = fx_1
    fy = fy_1

    G_temp = s

    while fin == 0:

        sum_max = 0  # small number for the metric (thresholding)

        arrayY = np.linspace(int(10 * (fy - step * G_temp)), int(10 * (fy + step * G_temp)), int(10 * step))
        arrayX = np.linspace(int(10 * (fx - step * G_temp)), int(10 * (fx + step * G_temp)), int(10 * step))

        for fx_temp in arrayX:
            for fy_temp in arrayY:

                fx_tmp, fy_tmp = fx_temp / 10, fy_temp / 10

                # print (fx_tmp, fy_tmp)

                theta_x = math.asin((fx_0 - fx_tmp) * wavelength / (N * dx))
                theta_y = math.asin((fy_0 - fy_tmp) * wavelength / (M * dy))
                ref_wave = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))

                # Compensation of the tilting angle for the off-axis acquisition
                reconstruction = holo_filter * ref_wave
                # module 2-pi phase retrieval
                phase = np.angle(reconstruction)
                # Thresholding process
                minVal = np.amin(phase)
                maxVal = np.amax(phase)
                phase_sca = (phase - minVal) / (maxVal - minVal)
                binary_phase = (phase_sca > 0.2)

                # plt.imshow(phase, cmap='gray'), plt.title("binary_phase")  # image in gray scale
                # plt.show()  # show image

                # Applying the summation and thresholding metric
                sum = np.sum(np.sum(binary_phase))
                if (sum > sum_max):
                    x_max_out = fx_tmp
                    y_max_out = fy_tmp
                    sum_max = sum
                # print (sum)

        G_temp = G_temp - 1

        if x_max_out == fx and y_max_out == fy:
            fin = 1;

        fx = x_max_out
        fy = y_max_out

    # print(x_max_out, y_max_out)

    # Retrieving the best reconstruction (compensated phase)
    theta_x = math.asin((fx_0 - x_max_out) * wavelength / (N * dx))
    theta_y = math.asin((fy_0 - y_max_out) * wavelength / (M * dy))
    ref_wave = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    comp_phase = holo_filter * ref_wave

    print("Phase compensation finished.")

    return comp_phase


def CFS(inp, wavelength, dx, dy):
    '''
    # Function to compensate phase maps of image plane off-axis DHM via a cost-function search algorithm.
    # Inputs:
    # inp - The input intensity (captured) hologram
    # wavelength - Wavelength of the illumination source to register the DHM hologram
    # dx, dy - Pixel dimensions of the camera sensor used for recording the hologram
    '''
    
    if scipy.__version__ == None:
        print('Please, install scipy.optimize library and import minimize function')
        sys.exit()

    # determine if the hologram is off-axis
    orders, thresh = regime(inp)
    if orders < 3:
        print('CFS algorithm requires off-axis holograms')
        sys.exit()

    # Retrieving the input shape
    inp = np.array(inp)
    M, N = inp.shape

    # if M and N are not equals, cutting the hologram in a square shape
    if M > N:
        cut = round(int(M - N) / 2)
        inp = inp[cut:(M - cut):N, 0:N]
        M, N = inp.shape
        print('The hologram has been cut into a square shape of: [', M, 'x', N, ']')
    elif M < N:
        cut = round(int(N - M) / 2)
        inp = inp[0:M, cut:(N - cut)]
        M, N = inp.shape
        print('The hologram has been cut into a square shape of: [', M, 'x', N, ']')
    else:
        inp = inp
        M, N = inp.shape

    # Creating a mesh_grid to operate in world coordinates
    x = np.arange(0, M, 1)  # array x
    y = np.arange(0, N, 1)  # array y
    X, Y = np.meshgrid(x - (M / 2), y - (N / 2), indexing='xy')  # meshgrid XY
    fx_0 = M / 2
    fy_0 = N / 2
    k = (2 * math.pi) / wavelength

    # The spatial filtering process is executed
    print("Spatial filtering process started.....")
    holo_filter, fx_max, fy_max = spatialFilteringCF(inp, M, N)
    print("Spatial filtering process finished.")

    # loading seeds
    seeds = [fx_max, fy_max]

    # minimization
    print("Minimization process started.....")
    step = 1
    res = minimize(costFunction, seeds, args=(M, N, holo_filter, wavelength, dx, dy, X, Y, fx_0, fy_0, k),
                   method='TNC', bounds=((seeds[0] - step, seeds[0] + step), (seeds[1] - step, seeds[1] + step)), tol=1e-3)
    print("Minimization process finished.")
    fx_max = res.x[0]
    fy_max = res.x[1]
    print('x: ',fx_max)
    print('y: ',fy_max)
    
    # Best phase compensation
    print("Phase compensation started....")
    theta_x = math.asin((fx_0 - fx_max) * wavelength / (M * dx))
    theta_y = math.asin((fy_0 - fy_max) * wavelength / (N * dy))
    ref_wave = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    comp_phase = holo_filter * ref_wave

    print("Phase compensation finished.")

    return comp_phase


def CNT(inp, wavelength, dx, dy, x1=None, x2=None, y1=None, y2=None, spatialFilter=None):
    '''
    # Function to compensate phase maps of image plane off-axis DHM, operating in non-telecentric regimen
    # Inputs:
    # inp - The input intensity (captured) hologram
    # wavelength - Wavelength of the illumination source to register the DHM hologram
    # dx, dy - Pixel dimensions of the camera sensor used for recording the hologram
    # x1 - Coordinate x1 for the rectangle (upper left corner)
    # y1 - Coordinate y1 for the rectangle (upper left corner)
    # x2 - Coordinate x2 for the rectangle (lower right corner)
    # y2 - Coordinate y2 for the rectangle (lower right corner)
    # spatialFilter - The approach to compute the spatial filter, two options available      and sfmr
    '''

    wavelength = wavelength * 0.000001
    dx = dx * 0.000001
    dy = dy * 0.000001

    # Retrieving the input shape
    inp = np.array(inp)
    M, N = inp.shape
    k = (2 * math.pi) / wavelength

    # Creating a mesh-grid to operate in world coordinates
    x = np.arange(0, N, 1)  # array x
    y = np.arange(0, M, 1)  # array y
    X, Y = np.meshgrid(x - (N / 2), y - (M / 2), indexing='xy')  # meshgrid XY

    # The spatial filtering process is executed
    print("Spatial filtering process started.....")
    if x1 is None and x2 is None and y1 is None and y2 is None:
        if spatialFilter == 'sfmr':
            Xcenter, Ycenter, holo_filter, ROI_array = spatialFilterinCNT(inp, M, N)
        else:
            print("Please use the sfmr option for the spatial filter or insert the values for  (x1,y1,x2,y2) if the spatial filter option is ‘sfr’  ")
            sys.exit()
    else:
        if spatialFilter == 'sfr':
            Xcenter, Ycenter, holo_filter, ROI_array = spatialFilterinCNT_II(inp, M, N, x1, y1, x2, y2)
        else:
            print("Please use the sfr option for the spatial filter if (x1,y1,x2,y2) are used as inputs or remove the (x1,y1,x2,y2) as inputs of the CNT function")
            sys.exit()
    print("Spatial filtering process finished.")

    # Fourier transform to the hologram filtered
    ft_holo = FT(holo_filter)
    FT_display = intensity(ft_holo, False)
    #imageShow(FT_display, 'FT Filtered')

    # reference wave for the first compensation (global linear compensation)
    ThetaXM = math.asin((N / 2 - Xcenter) * wavelength / (M * dx))
    ThetaYM = math.asin((M / 2 - Ycenter) * wavelength / (N * dy))
    reference = np.exp(1j * k * (math.sin(ThetaXM) * X * dx + math.sin(ThetaYM) * Y * dy))

    # First compensation
    comp_phase = holo_filter * reference
    phase_c = phase(comp_phase)

    # show the first compensation
    minVal = np.amin(phase_c)
    maxVal = np.amax(phase_c)
    phase_normalized = (phase_c - minVal) / (maxVal - minVal)
    binary_phase = (phase_normalized > 0.2)
    imageShow(binary_phase, 'Binarized phase')


    # creating the new reference wave to eliminate the circular phase factors
    m = abs(ROI_array[2] - ROI_array[0])
    n = abs(ROI_array[3] - ROI_array[1])
    Cx = np.power((M * dx), 2)/(wavelength * m)
    Cy = np.power((N * dy), 2)/(wavelength * n)
    cur = (Cx + Cy)/2

    print("Carefully determine the center of the circular phase factor in the Binarized Image...")
    p = input("Enter the pixel position X_cent of the center of circular phase map on x axis ")
    q = input("Enter the pixel position Y_cent of the center of circular phase map on y axis ")
    f = ((M/2) - int(p))/2
    g = ((N/2) - int(q))/2
    print("Phase compensation started....")

    cont = 0
    sum_max = 0
    s = 100
    step = 50
    perc = 40/100

    arrayCurvature = np.arange(cur - (cur*perc), cur + (cur*perc), perc/6)
    arrayXcenter = np.arange(f - s, f + s, step)
    arrayYcenter = np.arange(g - s, g + s, step)
    for curTemp in arrayCurvature:
        for fTemp in arrayXcenter:
            for gTemp in arrayYcenter:
                cont = cont + 1
                phi_spherical = (np.power(X - fTemp, 2) * np.power(dx, 2) / curTemp) + (
                np.power(Y - gTemp, 2) * np.power(dy, 2) / curTemp)
                phi_spherical = math.pi * phi_spherical / wavelength
                phi_spherical = np.exp(-1j * phi_spherical)

                phaseCompensate = comp_phase * phi_spherical
                phaseCompensate = np.angle(phaseCompensate)
                #imageShow(phaseCompensate, 'phaseCompensate')

                minVal = np.amin(phaseCompensate)
                maxVal = np.amax(phaseCompensate)
                phase_sca = (phaseCompensate - minVal) / (maxVal - minVal)
                binary_phase = (phase_sca > 0.2)
                #imageShow(binary_phase, 'phaseCompensate')

                # Applying the summation and thresholding metric
                sum = np.sum(np.sum(binary_phase))
                if (sum > sum_max):
                    f_out = fTemp
                    g_out = gTemp
                    cur_out = curTemp
                    sum_max = sum

    #print("after first search ", f_out, g_out, cur_out)

    cont = 0
    sum_max = 0
    s = 10
    step = 2
    perc = 0.1
    arrayXcenter = np.arange(f_out - s, f_out + s, step)
    arrayYcenter = np.arange(g_out - s, g_out + s, step)
    arrayCurvature = np.arange(cur_out - (cur_out*perc), cur_out + (cur_out*perc), 0.01)
    #arrayCurvature = np.arange(1.003, 1.03, 0.01)

    for curTemp in arrayCurvature:
        for fTemp in arrayXcenter:
            for gTemp in arrayYcenter:
                #print(curTemp)
                cont = cont + 1
                phi_spherical = (np.power(X - fTemp, 2) * np.power(dx, 2) / curTemp) + (
                    np.power(Y - gTemp, 2) * np.power(dy, 2) / curTemp)
                phi_spherical = math.pi * phi_spherical / wavelength
                phi_spherical = np.exp(-1j * phi_spherical)

                phaseCompensate = comp_phase * phi_spherical
                phaseCompensate = np.angle(phaseCompensate)
                #imageShow(phaseCompensate, 'phaseCompensate')

                minVal = np.amin(phaseCompensate)
                maxVal = np.amax(phaseCompensate)
                phase_sca = (phaseCompensate - minVal) / (maxVal - minVal)
                binary_phase = (phase_sca > 0.2)
                #imageShow(binary_phase, 'phaseCompensate')

                # Applying the summation and thresholding metric
                sum = np.sum(np.sum(binary_phase))
                #print(sum, curTemp)
                if (sum > sum_max):
                    f_out = fTemp
                    g_out = gTemp
                    cur_out = curTemp
                    sum_max = sum

    phi_spherical = (np.power(X - f_out, 2) * np.power(dx, 2) / cur_out) + (
            np.power(Y - g_out, 2) * np.power(dy, 2) / cur_out)
    phi_spherical = math.pi * phi_spherical / wavelength
    phi_spherical = np.exp(-1j * phi_spherical)
    phaseCompensate = comp_phase * phi_spherical

    
    #print("after fine compensation", f_out, g_out, cur_out)

    print("Phase compensation finished.")

    return phaseCompensate

'''
Auxiliary functions
'''

# Function to determine if the holograms is off-axis or not
def regime(inp):
    holoFT = np.float32(inp)
    fft_holo = cv2.dft(holoFT, flags=cv2.DFT_COMPLEX_OUTPUT)
    fft_holo = np.fft.fftshift(fft_holo)
    fft_holo_image = 20 * np.log(cv2.magnitude(fft_holo[:, :, 0], fft_holo[:, :, 1]))
    minVal = np.amin(np.abs(fft_holo_image))
    maxVal = np.amax(np.abs(fft_holo_image))
    fft_holo_image = cv2.convertScaleAbs(fft_holo_image, alpha=255.0 / (maxVal - minVal),
                                         beta=-minVal * 255.0 / (maxVal - minVal))

    # cv2.imshow('Binary image_resize', fft_holo_image)
    # cv2.waitKey(0)

    # apply binary thresholding
    ret, thresh = cv2.threshold(fft_holo_image, 200, 255, cv2.THRESH_BINARY)
    #cv2.imshow('Binary image', thresh)
    thresh_rize = cv2.resize(thresh, (1024, 1024))
    #cv2.imshow('Binary image_resize', thresh_rize)
    #cv2.waitKey(0)


    contours, hierarchy = cv2.findContours(image=thresh_rize, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_SIMPLE)
    # draw contours on the original image
    fft_holo_image = cv2.resize(thresh, (1024, 1024))
    image_copy = fft_holo_image.copy()
    cv2.drawContours(image=image_copy, contours=contours, contourIdx=-1, color=(0, 255, 0), thickness=2,
                     lineType=cv2.LINE_AA)
    # cv2.imshow('None approximation', image_copy)
    # cv2.waitKey(0)
    orders = len(contours)
    # print(orders)
    return orders, thresh

# Spatial filtering process for rectangular selection for CNT
def spatialFilterinCNT_II(inp, M, N, x1, y1, x2, y2):
    ROI_array = np.zeros(4)

    # Fourier transform of the hologram
    FT = np.fft.fft2(inp)
    FT = np.fft.fftshift(FT)

    # Filter to find the X_center and Y_center
    mask = np.zeros((M, N))
    mask[y1:y2, x1:x2] = 1

    # computing the center of the rectangle mask (X_center and Y_center)
    Xcenter = x1 + (x2 - x1)/2
    Ycenter = y1 + (y2 - y1)/2

    # spatial filter
    filter = FT * mask
    holo_filter = np.fft.ifftshift(filter)
    holo_filter = np.fft.ifft2(holo_filter)

    # creating array for coordinates x1, y1, x2, y2
    ROI_array[0] = x1
    ROI_array[1] = y1
    ROI_array[2] = x1 + x2
    ROI_array[3] = y1 + y2

    return Xcenter, Ycenter, holo_filter, ROI_array

# Spatial filtering process - manual selection for CNT
def spatialFilterinCNT(inp, M, N):
    ROI_array = np.zeros(4)
    holoFT = np.float32(inp)  # convertion of data to float
    fft_holo = cv2.dft(holoFT, flags=cv2.DFT_COMPLEX_OUTPUT)  # FFT of hologram
    fft_holo = np.fft.fftshift(fft_holo)
    fft_holo_image = 20 * np.log(cv2.magnitude(fft_holo[:, :, 0], fft_holo[:, :, 1]))  # logaritm scale FFT
    minVal = np.amin(np.abs(fft_holo_image))
    maxVal = np.amax(np.abs(fft_holo_image))
    fft_holo_image = cv2.convertScaleAbs(fft_holo_image, alpha=255.0 / (maxVal - minVal),
                                         beta=-minVal * 255.0 / (maxVal - minVal))

    ROI = cv2.selectROI(fft_holo_image, fromCenter=True)  # module to  ROI
    # imCrop = fft_holo_image[int(ROI[1]):int(ROI[1] + ROI[3]), int(ROI[0]):int(ROI[0] + ROI[2])]
    x1_ROI = int(ROI[1])
    y1_ROI = int(ROI[0])
    x2_ROI = int(ROI[1] + ROI[3])
    y2_ROI = int(ROI[0] + ROI[2])
    ROI_array[0] = x1_ROI
    ROI_array[1] = y1_ROI
    ROI_array[2] = x2_ROI
    ROI_array[3] = y2_ROI

    # computing the center of the rectangle mask
    Ycenter = x1_ROI + (x2_ROI - x1_ROI)/2
    Xcenter = y1_ROI + (y2_ROI - y1_ROI)/2

    holo_filter = np.zeros((M, N, 2))
    holo_filter[x1_ROI:x2_ROI, y1_ROI: y2_ROI] = 1
    holo_filter = holo_filter * fft_holo
    holo_filter = np.fft.ifftshift(holo_filter)
    holo_filter = cv2.idft(holo_filter, flags=cv2.DFT_INVERSE)

    holo_filter_real = holo_filter[:, :, 0]
    holo_filter_imag = holo_filter[:, :, 1]
    holo_filter = np.zeros((M, N), complex)
    for p in range(M):
        for q in range(N):
            holo_filter[p, q] = complex(holo_filter_real[p, q], holo_filter_imag[p, q])

    return Xcenter, Ycenter, holo_filter, ROI_array

# Function to display an image
def imageShow(inp, title):
    plt.imshow(inp, cmap='gray'), plt.title(title)  # image in gray scale
    plt.show()  # show image

    return

# Function to calculate the intensity representation of a given complex field
def intensity(inp, log):
    # Inputs:
    # inp - The input complex field
    # log - boolean variable to determine if a log representation is applied

    out = np.abs(inp)
    out = out * out

    if log == True:
        out = 20 * np.log(out)

    return out

# Function to calcule the amplitude representation of a given complex field
def amplitude(inp, log):
    out = np.abs(inp)

    if log == True:
        out = 20 * np.log(out)

    return out

# Function to calcule the phase representation of a given complex field using the
def phase(inp):
    out = np.angle(inp)

    return out
    
# Function to calcule the Fourier transform of a given field using the
def FT(inp):
    FT = np.fft.fft2(inp)
    FT = np.fft.fftshift(FT)

    return FT

# Spatial filtering for the FRS and ERS algorithms
def spatialFiltering(inp, M, N, upper):
    field_spec = np.fft.fftshift(inp)
    field_spec = np.fft.fft2(field_spec)
    field_spec = np.fft.fftshift(field_spec)

    pow_espec = intensity(field_spec, True)

    # Defining width and hieght of point of maxima in the ROI determination
    # 10% is estimated to be the proper distance to separated an in-line hologram from an off-axis one in the Spectrum
    M_search = int(M - (M * 0.1))
    N_search = int(N - (N * 0.1))
    # print (M_search)

    # Location and size of the ROI for the +1 or -1 diffraction order info

    # Location of the center of the ROI in the DHM hologram spectrum
    if (upper == True):
        # Find the location of the maxima point in the upper part of the DHM hologram spectrum
        indI = np.unravel_index(np.argmax(pow_espec[1:(int(M / 2) - M_search), 1:N], axis=None),
                                pow_espec[1:(int(M / 2) - M_search), 1:N].shape)
        # print (indI)
    else:
        # Find the location of the maxima point in the lower part of the DHM hologram spectrum
        indI = np.unravel_index(np.argmax(pow_espec[1:M, 1:(int(N / 2) - N_search)], axis=None),
                                pow_espec[1:M, 1:(int(N / 2) - N_search)].shape)
        # print (indI)

    # Determination of the ROI size. To do this, we use the theoretical size of the +1 or -1 diffraction orders according...
    # ... to the distance of their centers to the DC coordinates (middle point in the DHM hologram spectrum).
    d = np.sqrt(np.power(indI[1] - M / 2, 2) + np.power(indI[0] - N / 2, 2))
    radius = d / 3

    # Mask building
    mask = np.zeros((M, N))
    for j in range(M):
        for i in range(N):
            if np.power(j - indI[1], 2) + np.power(i - indI[0], 2) < np.power(radius, 2):
                mask[i, j] = 1

    tmp = field_spec * mask
    # field_spec.fill(0)

    # pow_espec = intensity(tmp, True)
    # out = pow_espec

    # Coming back to spatial domain (retreiving filtered hologram)
    out = np.fft.ifftshift(tmp)
    out = np.fft.ifft2(out)
    out = np.fft.ifftshift(out)

    return out, indI, tmp


# Spatial filtering process for FCF implementation
def spatialFilteringCF(inp, M, N):
    field_spec = np.fft.fft2(inp)
    field_spec = np.fft.fftshift(field_spec)

    pow_espec = intensity(field_spec, True)

    # Finding the max peaks for +1 order in I or II quadrant
    height_half = round(int(M / 2))
    mask = np.zeros((M, N))
    mask[0:height_half - 1, 0:M] = 1
    field_spec_tem = pow_espec * mask
    maximum = np.amax(field_spec_tem)
    fy_max, fx_max = np.where(field_spec_tem == maximum)
    #print(fx_max, fy_max )

    # Determination of the ROI size. To do this, we use the theoretical size of the +1 or -1 diffraction orders according...
    # ... to the distance of their centers to the DC coordinates (middle point in the DHM hologram spectrum).
    d = np.sqrt(np.power(fx_max - M / 2, 2) + np.power(fy_max - N / 2, 2))
    radius = d / 3

    # Mask building
    mask = np.zeros((M, N), dtype=int)
    for p in range(0, N):
        for q in range(0, M):
            if np.sqrt((p - fy_max) ** 2 + (q - fx_max) ** 2) < radius:
                mask[p, q] = 1

    # Filtering the hologram
    tmp = field_spec * mask

    # Coming back to spatial domain (retrieving filtered hologram)
    out = np.fft.ifftshift(tmp)
    out = np.fft.ifft2(out)

    return out, fx_max, fy_max


# cost function for the CFS implementation
def costFunction(seeds, M, N, holo_filter, wavelength, dx, dy, X, Y, fx_0, fy_0, k):
    J = 0;
    theta_x = math.asin((fx_0 - seeds[0]) * wavelength / (M * dx))
    theta_y = math.asin((fy_0 - seeds[1]) * wavelength / (N * dy))
    ref_wave = np.exp(1j * k * ((math.sin(theta_x) * X * dx) + (math.sin(theta_y) * Y * dy)))
    phase = compensation(holo_filter, ref_wave)

    phase = phase + math.pi
    sumIB = phase.sum()
    J = (M * N) - sumIB

    return J


# function for the compensation
def compensation(holo_filter, ref_wave):
    reconstruction = holo_filter * ref_wave
    phase = np.angle(reconstruction)

    return phase
    



"""
Title-->            Numerical propagator package
Author-->           Carlos Trujillo, Ana Doblas and Raul Castaneda
Date-->             21/06/2021
Last modified-->    04/11/2022
Groups-->           University of Memphis -  Optical Imaging Research laboratory (OIRL)
                    EAFIT University - Applied Optics Group
Abstract -->        Script for implementing three numerical propagators via the pyDHM library
                    Angular spectrum, Fresnel transform, and Fresnel-Bluestein transform
Links-->            More information can be found on the website
                    https://github.com/catrujilla/pyDHM
"""

# libraries
import numpy as np
from math import pi


def fresnel(field, z, wavelength, dx, dy):
    """
    # Function to diffract a complex field using Fresnel approximation with Fourier method
    # Inputs:
    # field - complex field
    # z - propagation distance
    # wavelength - wavelength
    # dx, dy - sampling pitches
    """

    field = np.array(field)
    M, N = field.shape
    x = np.arange(0, N, 1)  # array x
    y = np.arange(0, M, 1)  # array y
    X, Y = np.meshgrid(x - (N / 2), y - (M / 2), indexing='xy')

    dxout = (wavelength * z) / (M * dx)
    dyout = (wavelength * z) / (N * dy)

    z_phase = np.exp2(1j * 2 * pi * z / wavelength) / (1j * wavelength * z)

    out_phase = np.exp2((1j * pi / (wavelength * z)) * (np.power(X * dxout, 2) + np.power(Y * dyout, 2)))
    in_phase = np.exp2((1j * pi / (wavelength * z)) * (np.power(X * dx, 2) + np.power(Y * dy, 2)))

    tmp = (field * in_phase)
    tmp = np.fft.fftshift(tmp)
    tmp = np.fft.fft2(tmp)
    tmp = np.fft.fftshift(tmp)

    out = z_phase * out_phase * dx * dy * tmp

    return out


def bluestein(field, z, wavelength, dx, dy, dxout, dyout):
    """
    # Function to diffract a complex field using Fresnel-Bluesteins approximation
    # Inputs:
    # field - complex field
    # z - propagation distance
    # lambda - wavelength
    # dx, dy - sampling pitches
    # dxout, dyout - output window pitches
    """

    field = np.array(field)
    M, N = field.shape
    x = np.arange(0, N, 1)  # array x x = np.arange(0, N, 1)
    y = np.arange(0, M, 1)  # array y y = np.arange(0, M, 1)
    X, Y = np.meshgrid(x - (N / 2), y - (M / 2), indexing='xy')

    padx = int(M / 2)
    pady = int(N / 2)

    z_phase = np.exp2(1j * 2 * pi * z / wavelength) / (1j * wavelength * z)

    output_phase = np.exp2(
        (-1j * pi / (wavelength * z)) * (dxout * (dx - dxout) * np.power(X, 2) + dyout * (dy - dyout) * np.power(Y, 2)))

    input_phase1 = np.exp2(
        ((1j * pi) / (wavelength * z)) * (dx * (dx - dxout) * np.power(X, 2) + dy * (dy - dyout) * np.power(Y, 2)))
    input_phase2 = np.exp2(((1j * pi) / (wavelength * z)) * (dx * dxout * np.power(X, 2) + dy * dyout * np.power(Y, 2)))

    f1 = np.pad(field * input_phase1, ((padx, padx), (pady, pady)), mode='constant')
    IP1 = np.fft.fftshift(f1)
    IP1 = np.fft.fft2(IP1)
    IP1 = np.fft.fftshift(IP1)

    f2 = np.pad(input_phase2, ((padx, padx), (pady, pady)), mode='constant')
    IP2 = np.fft.fftshift(f2)
    IP2 = np.fft.fft2(IP2)
    IP2 = np.fft.fftshift(IP2)

    out = np.fft.ifftshift(IP1 * IP2)
    out = np.fft.ifft2(out)
    out = np.fft.ifftshift(out)

    out = out[padx:padx + M, pady:pady + N]
    out = out * z_phase
    out = out * output_phase

    return out


def angularSpectrum(field, z, wavelength, dx, dy):
    '''
    # Function to diffract a complex field using the angular spectrum approximation
    # Inputs:
    # field - complex field
    # z - propagation distance
    # wavelength - wavelength
    # dx,dy - sampling pitches
    '''

    field = np.array(field)
    M, N = field.shape
    x = np.arange(0, N, 1)  # array x
    y = np.arange(0, M, 1)  # array y
    X, Y = np.meshgrid(x - (N / 2), y - (M / 2), indexing='xy')

    dfx = 1 / (dx * M)
    dfy = 1 / (dy * N)

    field_spec = np.fft.fftshift(field)
    field_spec = np.fft.fft2(field_spec)
    field_spec = np.fft.fftshift(field_spec)

    phase = np.exp2(1j * z * pi * np.sqrt(np.power(1 / wavelength, 2) - (np.power(X * dfx, 2) + np.power(Y * dfy, 2))))

    tmp = field_spec * phase

    out = np.fft.ifftshift(tmp)
    out = np.fft.ifft2(out)
    out = np.fft.ifftshift(out)

    return out




import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text(encoding="utf8")

# This call to setup() does all the work
setup(
    name="pyDHM",
    version="1.0.4",
    description="An open-source Python library to numerically recover the complex wavefield information of samples from Digital Holographic Microscopy (DHM)",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/catrujilla/pyDHM",
    author="Carlos Trujillo",
    author_email="catrujilla@eafit.edu.co",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["pyDHM"],
    include_package_data=True,
    install_requires=["matplotlib", "scipy"],
)


# -*- coding: utf-8 -*-
"""
Title-->            Utility script
Author-->           Ana Doblas, Carlos Trujillo, Raul Castaneda,
Date-->             03/03/2019
Last modified-->    16/04/2022
                    University of Memphis
                    Optical Imaging Research lab (OIRL)
                    EAFIT University
                    Applied Optics Group
Abstract -->        Script that implements the different methods to render the resulting complex field data
Links-->          - https://unal-optodigital.github.io/JDiffraction/
"""

import cv2
import sys
import numpy as np
from matplotlib import pyplot as plt
from PIL import Image, ImageOps
from scipy import ndimage

# circular spatial filter
def sfc(field, radius, centX, centY, display):
    """
    # Function to create a spatial filter with a circle mask
    # Inputs:
    # field - The field to be filtered
    # radius - dimension of the radius for the circle
    # centX - center circle position in x axis
    # centY - center circle position in y axis
    # display - boolean variable to visualize the FT spatial filter

    """
    field = np.array(field)
    M, N = field.shape

    circle = np.zeros((radius * 2, radius * 2), dtype=int)
    for p in range(0, radius * 2):
        for q in range(0, radius * 2):
            if np.sqrt((p - radius) ** 2 + (q - radius) ** 2) < radius:
                circle[p, q] = 1

    (minC, maxC) = circle.shape
    mask = np.zeros((M, N), dtype=int)
    minX = int(centX - minC / 2)
    maxX = int(centX + minC / 2)
    minY = int(centY - minC / 2)
    maxY = int(centY + maxC / 2)
    mask[minY:maxY, minX:maxX] = circle[:, :]

    FT = np.fft.fft2(field)
    FT = np.fft.fftshift(FT)
    filter = FT * mask

    if (display == True):
        holo_filter_display = intensity(filter, True)
        imageShow(holo_filter_display, 'Spatial filter')

    min1X = int(N / 2 - (maxX - minX) / 2)
    max1X = int(N / 2 + (maxX - minX) / 2)
    min1Y = int(M / 2 - (maxY - minY) / 2)
    max1Y = int(M / 2 + (maxY - minY) / 2)

    crop_real = np.zeros((M, N))
    crop_real[min1Y:max1Y, min1X:max1X] = filter.real[minY:maxY, minX:maxX]

    crop_imag = np.zeros((M, N))
    crop_imag[min1Y:max1Y, min1X:max1X] = filter.imag[minY:maxY, minX:maxX]

    crop = crop_real + 1j * crop_imag
    ift = np.fft.ifftshift(crop)

    holoFilter = np.fft.ifft2(ift)
    field = holoFilter

    return field

# rectangular spatial filter
def sfr(field, x1, x2, y1, y2, display):
    """
    # Function to create a spatial filter with a rectangle mask
    # Inputs:
    # field - The field to be filtered
    # x1 - Coordinate x1 for the rectangle (upper left corner)
    # y1 - Coordinate y1 for the rectangle (upper left corner)
    # x2 - Coordinate x2 for the rectangle (lower right corner)
    # y2 - Coordinate y2 for the rectangle (lower right corner)
    # display - boolean variable (True or False) to visualize the FT spatial filter
    """
    field = np.array(field)
    M, N = field.shape

    mask = np.zeros((M, N))
    mask[y1:y2, x1:x2] = 1
    FT = np.fft.fft2(field)
    FT = np.fft.fftshift(FT)
    filter = FT * mask

    if (display == True):
        holo_filter_display = intensity(filter, True)
        imageShow(holo_filter_display, 'Spatial filter')

    minX = int(N / 2 - (x2 - x1) / 2)
    maxX = int(N / 2 + (x2 - x1) / 2)
    minY = int(M / 2 - (y2 - y1) / 2)
    maxY = int(M / 2 + (y2 - y1) / 2)

    crop_real = np.zeros((M, N))
    crop_real[minY:maxY, minX:maxX] = filter.real[y1:y2, x1:x2]
    crop_imag = np.zeros((M, N))
    crop_imag[minY:maxY, minX:maxX] = filter.imag[y1:y2, x1:x2]

    crop = crop_real + 1j * crop_imag
    ift = np.fft.ifftshift(crop)

    holoFilter = np.fft.ifft2(ift)
    field = holoFilter

    return field

# manual rectangular spatial filter
def sfmr(field, display):
    """
    # Function to create a spatial filter with a rectangle mask
    # Inputs:
    # field - The field to be filtered
    # display - boolean variable (True or False) to visualize the FT spatial filter
    """
        
    if cv2.__version__ == None:
        print('Please install OpenCV library before using the sfmr function')
        sys.exit()

    field = np.array(field)
    M, N = field.shape

    ROI_arrray = np.zeros(4)
    holoFT = np.float32(field)
    fft_holo = cv2.dft(holoFT, flags=cv2.DFT_COMPLEX_OUTPUT)
    fft_holo = np.fft.fftshift(fft_holo)
    fft_holo_image = 20 * np.log(cv2.magnitude(fft_holo[:, :, 0], fft_holo[:, :, 1]))
    minVal = np.amin(np.abs(fft_holo_image))
    maxVal = np.amax(np.abs(fft_holo_image))
    fft_holo_image = cv2.convertScaleAbs(fft_holo_image, alpha=255.0 / (maxVal - minVal),
                                         beta=-minVal * 255.0 / (maxVal - minVal))

    ROI = cv2.selectROI(fft_holo_image, fromCenter=True)
    x1_ROI = int(ROI[1])
    y1_ROI = int(ROI[0])
    x2_ROI = int(ROI[1] + ROI[3])
    y2_ROI = int(ROI[0] + ROI[2])
    ROI_arrray[0] = x1_ROI
    ROI_arrray[1] = y1_ROI
    ROI_arrray[2] = x2_ROI
    ROI_arrray[3] = y2_ROI

    holo_filter = np.zeros((M, N, 2))
    holo_filter[x1_ROI:x2_ROI, y1_ROI: y2_ROI] = 1
    holo_filter = holo_filter * fft_holo

    if (display == True):
        holo_filter_real = holo_filter[:, :, 0]
        holo_filter_imag = holo_filter[:, :, 1]
        holo_filter_display = np.zeros((M, N), complex)
        for p in range(M):
            for q in range(N):
                holo_filter_display[p, q] = complex(holo_filter_real[p, q], holo_filter_imag[p, q])

        holo_filter_display = intensity(holo_filter_display, True)
        imageShow(holo_filter_display, 'Spatial filter')

    holo_filter = np.fft.ifftshift(holo_filter)
    holo_filter = cv2.idft(holo_filter, flags=cv2.DFT_INVERSE)

    holo_filter_real = holo_filter[:, :, 0]
    holo_filter_imag = holo_filter[:, :, 1]
    holo_filter = np.zeros((M, N), complex)
    for p in range(M):
        for q in range(N):
            holo_filter[p, q] = complex(holo_filter_real[p, q], holo_filter_imag[p, q])

    # return Xcenter, Ycenter, holo_filter, ROI_arrray
    return holo_filter

# Function to determine if the holograms is off-axis or not
def regime(inp):
    holoFT = np.float32(inp)
    fft_holo = cv2.dft(holoFT, flags=cv2.DFT_COMPLEX_OUTPUT)
    fft_holo = np.fft.fftshift(fft_holo)
    fft_holo_image = 20 * np.log(cv2.magnitude(fft_holo[:, :, 0], fft_holo[:, :, 1])) 
    minVal = np.amin(np.abs(fft_holo_image))
    maxVal = np.amax(np.abs(fft_holo_image))
    fft_holo_image = cv2.convertScaleAbs(fft_holo_image, alpha=255.0 / (maxVal - minVal),
                                         beta=-minVal * 255.0 / (maxVal - minVal))

    # apply binary thresholding
    ret, thresh = cv2.threshold(fft_holo_image, 200, 255, cv2.THRESH_BINARY)
    #cv2.imshow('Binary image', thresh)
    thresh_rize = cv2.resize(thresh, (1024, 1024))
    #cv2.imshow('Binary image_resize', thresh_rize)
    #cv2.waitKey(0)


    contours, hierarchy = cv2.findContours(image=thresh_rize, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_SIMPLE)
    # draw contours on the original image
    fft_holo_image = cv2.resize(thresh, (1024, 1024))
    image_copy = fft_holo_image.copy()
    cv2.drawContours(image=image_copy, contours=contours, contourIdx=-1, color=(0, 255, 0), thickness=2,
                     lineType=cv2.LINE_AA)
    #cv2.imshow('None approximation', image_copy)
    #cv2.waitKey(0)
    #print(len(contours))
    return ret, thresh
    
def HM2F(inp, kernel, figures, plots):
    if kernel % 2 == 0:
        print('Kernel size must be a odd number')
        exit()

    mean_image = inp
    cont = 1
    for i in range(3, kernel + 2, 2):
        filter = ndimage.median_filter(inp, i, mode='constant', cval=0)
        mean_image = (mean_image + filter) / 2
        imageDenoise = mean_image

    return imageDenoise


# Salutation function of the library
def salutation():
    print("Hello world! This is pyDHM library version 1.0")
    return


# Function to read an image file from the disk
def imageRead(namefile):
    Im = Image.open(namefile)
    loadImage = ImageOps.grayscale(Im)

    return loadImage


# Function to display an image
# Inputs:
# inp - The input complex field
# title - The title of the displayed image
def imageShow(inp, title):
    plt.imshow(inp, cmap='gray'), plt.title(title)  # image in gray scale

    if title == None:
        plt.imshow(inp, cmap='gray')
    plt.show()  # show image

    return
    

# Function to calcule the amplitude representation of a given complex field
# Inputs:
# inp - The input complex field
# log - boolean variable to determine if a log representation is applied
def amplitude(inp, log):
        
    out = np.abs(inp)
    if log == False or log == None:
        out
    else:
         out = 20 * np.log(out)

    return out


# Function to calcule the intensity representation of a given complex field
# Inputs:
# inp - The input complex field
# log - boolean variable to determine if a log representation is applied
def intensity(inp, log):
    out = np.abs(inp)
    if log == False or log == None:
        out = out * out
    else:
        out = 20 * np.log(out)
        out[out == np.inf] = 0
        out[out == -np.inf] = 0

    return out


# Function to calcule the phase representation of a given complex field using the
# function 'angle'
# Inputs:
# inp - The input complex field
def phase(inp):
    out = np.angle(inp)

    return out


# Function to calcule the Fourier transform of a given field using the
# 'fft of numpy'
# Inputs:
# inp - The input field
def FT(inp):
    FT = np.fft.fft2(inp)
    FT = np.fft.fftshift(FT)

    return FT

    
